"""
common parts of the test suite for rapidfuzz
"""

from __future__ import annotations

from dataclasses import dataclass
from math import isnan
from typing import Any

import pytest

from rapidfuzz import process_cpp, process_py

try:
    from pandas import NA as pandas_NA
except BaseException:
    pandas_NA = None


def _get_scorer_flags_py(scorer: Any, scorer_kwargs: dict[str, Any]) -> tuple[int, int]:
    params = getattr(scorer, "_RF_ScorerPy", None)
    if params is not None:
        flags = params["get_scorer_flags"](**scorer_kwargs)
        return (flags["worst_score"], flags["optimal_score"])
    return (0, 100)


def is_none(s):
    if s is None or s is pandas_NA:
        return True

    if isinstance(s, float) and isnan(s):
        return True

    return False


def call_and_maybe_catch(call, *args, catch_exceptions=False, **kwargs):
    if not catch_exceptions:
        return call(*args, **kwargs)

    try:
        return call(*args, **kwargs)
    except AssertionError as e:
        raise e
    except Exception as e:
        return e


def compare_exceptions(e1, e2):
    try:
        return str(e1) == str(e2)
    except Exception:
        return False


def scorer_tester(scorer, s1, s2, catch_exceptions=False, **kwargs):
    score1 = call_and_maybe_catch(scorer, s1, s2, **kwargs)
    exception = isinstance(score1, Exception)

    temp_kwargs = kwargs.copy()
    process_kwargs = {}

    if "processor" in kwargs:
        process_kwargs["processor"] = kwargs["processor"]
        del temp_kwargs["processor"]

    if "score_cutoff" in kwargs:
        process_kwargs["score_cutoff"] = kwargs["score_cutoff"]
        del temp_kwargs["score_cutoff"]

    if temp_kwargs:
        process_kwargs["scorer_kwargs"] = temp_kwargs

    extractOne_res1 = call_and_maybe_catch(
        process_cpp.extractOne, s1, [s2], catch_exceptions=catch_exceptions, scorer=scorer, **process_kwargs
    )
    extractOne_res2 = call_and_maybe_catch(
        process_py.extractOne, s1, [s2], catch_exceptions=catch_exceptions, scorer=scorer, **process_kwargs
    )
    extract_res1 = call_and_maybe_catch(
        process_cpp.extract, s1, [s2], catch_exceptions=catch_exceptions, scorer=scorer, **process_kwargs
    )
    extract_res2 = call_and_maybe_catch(
        process_py.extract, s1, [s2], catch_exceptions=catch_exceptions, scorer=scorer, **process_kwargs
    )
    extract_iter_res1 = call_and_maybe_catch(
        list, process_cpp.extract_iter(s1, [s2], scorer=scorer, **process_kwargs), catch_exceptions=catch_exceptions
    )
    extract_iter_res2 = call_and_maybe_catch(
        list, process_py.extract_iter(s1, [s2], scorer=scorer, **process_kwargs), catch_exceptions=catch_exceptions
    )

    if exception:
        assert compare_exceptions(extractOne_res1, score1)
        assert compare_exceptions(extractOne_res2, score1)
        assert compare_exceptions(extract_res1, score1)
        assert compare_exceptions(extract_res2, score1)
        assert compare_exceptions(extract_iter_res1, score1)
        assert compare_exceptions(extract_iter_res2, score1)
    elif is_none(s1) or is_none(s2):
        assert extractOne_res1 is None
        assert extractOne_res2 is None
        assert extract_res1 == []
        assert extract_res2 == []
        assert extract_iter_res1 == []
        assert extract_iter_res2 == []
    elif kwargs.get("score_cutoff") is not None:
        worst_score, optimal_score = _get_scorer_flags_py(scorer, process_kwargs.get("scorer_kwargs", {}))
        lowest_score_worst = optimal_score > worst_score
        is_filtered = score1 < kwargs["score_cutoff"] if lowest_score_worst else score1 > kwargs["score_cutoff"]

        if is_filtered:
            assert extractOne_res1 is None
            assert extractOne_res2 is None
            assert extract_res1 == []
            assert extract_res2 == []
            assert extract_iter_res1 == []
            assert extract_iter_res2 == []
        else:
            assert pytest.approx(score1) == extractOne_res1[1]
            assert pytest.approx(score1) == extractOne_res2[1]
            assert pytest.approx(score1) == extract_res1[0][1]
            assert pytest.approx(score1) == extract_res2[0][1]
            assert pytest.approx(score1) == extract_iter_res1[0][1]
            assert pytest.approx(score1) == extract_iter_res2[0][1]
    else:
        assert pytest.approx(score1) == extractOne_res1[1]
        assert pytest.approx(score1) == extractOne_res2[1]
        assert pytest.approx(score1) == extract_res1[0][1]
        assert pytest.approx(score1) == extract_res2[0][1]
        assert pytest.approx(score1) == extract_iter_res1[0][1]
        assert pytest.approx(score1) == extract_iter_res2[0][1]

    try:
        import numpy as np
    except Exception:
        np = None

    if np is not None:
        cdist_scores1 = call_and_maybe_catch(
            process_cpp.cdist, [s1], [s2], catch_exceptions=catch_exceptions, scorer=scorer, **process_kwargs
        )
        cdist_scores2 = call_and_maybe_catch(
            process_py.cdist, [s1], [s2], catch_exceptions=catch_exceptions, scorer=scorer, **process_kwargs
        )
        # probably trigger multi match / simd implementations
        cdist_scores3 = call_and_maybe_catch(
            process_cpp.cdist, [s1] * 2, [s2] * 4, catch_exceptions=catch_exceptions, scorer=scorer, **process_kwargs
        )
        cdist_scores4 = call_and_maybe_catch(
            process_py.cdist, [s1] * 2, [s2] * 4, catch_exceptions=catch_exceptions, scorer=scorer, **process_kwargs
        )

        if exception:
            assert compare_exceptions(cdist_scores1, score1)
            assert compare_exceptions(cdist_scores2, score1)
            assert compare_exceptions(cdist_scores3, score1)
            assert compare_exceptions(cdist_scores4, score1)
        else:
            assert np.all(np.isclose(cdist_scores1, score1))
            assert np.all(np.isclose(cdist_scores2, score1))
            assert np.all(np.isclose(cdist_scores3, score1))
            assert np.all(np.isclose(cdist_scores4, score1))

    if exception:
        raise score1

    return score1


def symmetric_scorer_tester(scorer, s1, s2, catch_exceptions=False, **kwargs):
    score1 = call_and_maybe_catch(scorer_tester, scorer, s1, s2, catch_exceptions=catch_exceptions, **kwargs)
    score2 = call_and_maybe_catch(scorer_tester, scorer, s2, s1, catch_exceptions=catch_exceptions, **kwargs)

    if isinstance(score1, Exception):
        assert compare_exceptions(score1, score2)
        raise score1

    assert pytest.approx(score1) == score2
    return score1


@dataclass
class Scorer:
    distance: Any
    similarity: Any
    normalized_distance: Any
    normalized_similarity: Any
    editops: Any
    opcodes: Any


class GenericScorer:
    def __init__(self, py_scorers, cpp_scorers, get_scorer_flags):
        self.py_scorers = py_scorers
        self.cpp_scorers = cpp_scorers
        self.scorers = self.py_scorers + self.cpp_scorers

        def validate_attrs(func1, func2):
            assert hasattr(func1, "_RF_ScorerPy")
            assert hasattr(func2, "_RF_ScorerPy")
            assert func1.__name__ == func2.__name__
            assert func1.__qualname__ == func2.__qualname__
            assert func1.__doc__ == func2.__doc__

        for scorer in self.scorers:
            validate_attrs(scorer.distance, self.scorers[0].distance)
            validate_attrs(scorer.similarity, self.scorers[0].similarity)
            validate_attrs(scorer.normalized_distance, self.scorers[0].normalized_distance)
            validate_attrs(scorer.normalized_similarity, self.scorers[0].normalized_similarity)

        for scorer in self.cpp_scorers:
            assert hasattr(scorer.distance, "_RF_Scorer")
            assert hasattr(scorer.similarity, "_RF_Scorer")
            assert hasattr(scorer.normalized_distance, "_RF_Scorer")
            assert hasattr(scorer.normalized_similarity, "_RF_Scorer")

        self.get_scorer_flags = get_scorer_flags

    def _editops(self, s1, s2, catch_exceptions=False, **kwargs):
        results = [
            call_and_maybe_catch(scorer.editops, s1, s2, catch_exceptions=catch_exceptions, **kwargs)
            for scorer in self.scorers
        ]

        for result in results:
            assert compare_exceptions(result, results[0])

        if any(isinstance(result, Exception) for result in results):
            raise results[0]

        return results[0]

    def _opcodes(self, s1, s2, catch_exceptions=False, **kwargs):
        results = [
            call_and_maybe_catch(scorer.opcodes, s1, s2, catch_exceptions=catch_exceptions, **kwargs)
            for scorer in self.scorers
        ]

        for result in results:
            assert compare_exceptions(result, results[0])

        if any(isinstance(result, Exception) for result in results):
            raise results[0]

        return results[0]

    def _distance(self, s1, s2, catch_exceptions=False, **kwargs):
        symmetric = self.get_scorer_flags(s1, s2, **kwargs)["symmetric"]
        tester = symmetric_scorer_tester if symmetric else scorer_tester

        scores = [
            call_and_maybe_catch(tester, scorer.distance, s1, s2, catch_exceptions=catch_exceptions, **kwargs)
            for scorer in self.scorers
        ]

        if any(isinstance(score, Exception) for score in scores):
            for score in scores:
                assert compare_exceptions(score, scores[0])
            raise scores[0]

        scores = sorted(scores)
        assert pytest.approx(scores[0]) == scores[-1]
        return scores[0]

    def _similarity(self, s1, s2, catch_exceptions=False, **kwargs):
        symmetric = self.get_scorer_flags(s1, s2, **kwargs)["symmetric"]
        tester = symmetric_scorer_tester if symmetric else scorer_tester

        scores = [
            call_and_maybe_catch(tester, scorer.similarity, s1, s2, catch_exceptions=catch_exceptions, **kwargs)
            for scorer in self.scorers
        ]

        if any(isinstance(score, Exception) for score in scores):
            for score in scores:
                assert compare_exceptions(score, scores[0])
            raise scores[0]

        scores = sorted(scores)
        assert pytest.approx(scores[0]) == scores[-1]
        return scores[0]

    def _normalized_distance(self, s1, s2, catch_exceptions=False, **kwargs):
        symmetric = self.get_scorer_flags(s1, s2, **kwargs)["symmetric"]
        tester = symmetric_scorer_tester if symmetric else scorer_tester

        scores = [
            call_and_maybe_catch(
                tester, scorer.normalized_distance, s1, s2, catch_exceptions=catch_exceptions, **kwargs
            )
            for scorer in self.scorers
        ]

        if any(isinstance(score, Exception) for score in scores):
            for score in scores:
                assert compare_exceptions(score, scores[0])
            raise scores[0]

        scores = sorted(scores)
        assert pytest.approx(scores[0]) == scores[-1]
        return scores[0]

    def _normalized_similarity(self, s1, s2, catch_exceptions=False, **kwargs):
        symmetric = self.get_scorer_flags(s1, s2, **kwargs)["symmetric"]
        tester = symmetric_scorer_tester if symmetric else scorer_tester

        scores = [
            call_and_maybe_catch(
                tester, scorer.normalized_similarity, s1, s2, catch_exceptions=catch_exceptions, **kwargs
            )
            for scorer in self.scorers
        ]

        if any(isinstance(score, Exception) for score in scores):
            for score in scores:
                assert compare_exceptions(score, scores[0])
            raise scores[0]

        scores = sorted(scores)
        assert pytest.approx(scores[0]) == scores[-1]
        return scores[0]

    def _validate(self, s1, s2, catch_exceptions=False, **kwargs):
        # todo requires more complex test handling
        # score_cutoff = kwargs.get("score_cutoff")
        kwargs = {k: v for k, v in kwargs.items() if k != "score_cutoff"}

        maximum = self.get_scorer_flags(s1, s2, **kwargs)["maximum"]

        dist = call_and_maybe_catch(self._distance, s1, s2, catch_exceptions=catch_exceptions, **kwargs)
        sim = call_and_maybe_catch(self._similarity, s1, s2, catch_exceptions=catch_exceptions, **kwargs)
        norm_dist = call_and_maybe_catch(self._normalized_distance, s1, s2, catch_exceptions=catch_exceptions, **kwargs)
        norm_sim = call_and_maybe_catch(
            self._normalized_similarity, s1, s2, catch_exceptions=catch_exceptions, **kwargs
        )

        if isinstance(dist, Exception):
            assert compare_exceptions(dist, sim)
            assert compare_exceptions(dist, norm_dist)
            assert compare_exceptions(dist, norm_sim)
            raise dist

        assert pytest.approx(dist) == maximum - sim
        if maximum != 0:
            assert pytest.approx(dist / maximum) == norm_dist
            assert pytest.approx(sim / maximum) == norm_sim
        else:
            assert pytest.approx(0.0) == norm_dist
            assert pytest.approx(1.0) == norm_sim

        return dist, sim, norm_dist, norm_sim

    def distance(self, s1, s2, catch_exceptions=False, **kwargs):
        dist, _, _, _ = self._validate(s1, s2, catch_exceptions=catch_exceptions, **kwargs)
        if "score_cutoff" not in kwargs:
            return dist

        return self._distance(s1, s2, catch_exceptions=catch_exceptions, **kwargs)

    def similarity(self, s1, s2, catch_exceptions=False, **kwargs):
        _, sim, _, _ = self._validate(s1, s2, catch_exceptions=catch_exceptions, **kwargs)
        if "score_cutoff" not in kwargs:
            return sim

        return self._similarity(s1, s2, catch_exceptions=catch_exceptions, **kwargs)

    def normalized_distance(self, s1, s2, catch_exceptions=False, **kwargs):
        if not is_none(s1) and not is_none(s2):
            _, _, norm_dist, _ = self._validate(s1, s2, catch_exceptions=catch_exceptions, **kwargs)
            # todo we should be able to handle this in a nicer way
            if "score_cutoff" not in kwargs:
                return norm_dist
        return self._normalized_distance(s1, s2, catch_exceptions=catch_exceptions, **kwargs)

    def normalized_similarity(self, s1, s2, catch_exceptions=False, **kwargs):
        if not is_none(s1) and not is_none(s2):
            _, _, _, norm_sim = self._validate(s1, s2, catch_exceptions=catch_exceptions, **kwargs)
            if "score_cutoff" not in kwargs:
                return norm_sim
        return self._normalized_similarity(s1, s2, catch_exceptions=catch_exceptions, **kwargs)

    def editops(self, s1, s2, catch_exceptions=False, **kwargs):
        editops_ = self._editops(s1, s2, catch_exceptions=catch_exceptions, **kwargs)
        opcodes_ = self._opcodes(s1, s2, catch_exceptions=catch_exceptions, **kwargs)
        assert opcodes_.as_editops() == editops_
        assert opcodes_ == editops_.as_opcodes()
        return editops_

    def opcodes(self, s1, s2, catch_exceptions=False, **kwargs):
        editops_ = self._editops(s1, s2, catch_exceptions=catch_exceptions, **kwargs)
        opcodes_ = self._opcodes(s1, s2, catch_exceptions=catch_exceptions, **kwargs)
        assert opcodes_.as_editops() == editops_
        assert opcodes_ == editops_.as_opcodes()
        return opcodes_



# SPDX-License-Identifier: MIT
# Copyright (C) 2025 Max Bachmann
# This file is generated by tools/generate_python.py
from __future__ import annotations

import contextlib
import os

from rapidfuzz._feature_detector import AVX2, SSE2, supports

__all__ = ["default_process"]

_impl = os.environ.get("RAPIDFUZZ_IMPLEMENTATION")
if _impl == "cpp":
    imported = False
    if supports(AVX2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.utils_cpp_avx2 import (
                default_process,  # pyright: ignore[reportMissingImports]
            )

            imported = True

    if not imported and supports(SSE2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.utils_cpp_sse2 import (
                default_process,  # pyright: ignore[reportMissingImports]
            )

            imported = True

    if not imported:
        from rapidfuzz.utils_cpp import (
            default_process,  # pyright: ignore[reportMissingImports]
        )
elif _impl == "python":
    from rapidfuzz.utils_py import default_process
else:
    imported = False
    if supports(AVX2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.utils_cpp_avx2 import (
                default_process,  # pyright: ignore[reportMissingImports]
            )

            imported = True

    if not imported and supports(SSE2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.utils_cpp_sse2 import (
                default_process,  # pyright: ignore[reportMissingImports]
            )

            imported = True

    if not imported:
        with contextlib.suppress(ImportError):
            from rapidfuzz.utils_cpp import (
                default_process,  # pyright: ignore[reportMissingImports]
            )

            imported = True

    if not imported:
        from rapidfuzz.utils_py import default_process



# distutils: language=c++
# cython: language_level=3, binding=True, linetrace=True

from libc.stdint cimport uint32_t


cdef extern from "FeatureDetector/CpuInfo.hpp":
    cdef int CPU_FEATURE_SSE2
    cdef int CPU_FEATURE_AVX2

    cdef cppclass CpuInfo:
        @staticmethod
        bint supports(uint32_t features)

SSE2 = CPU_FEATURE_SSE2
AVX2 = CPU_FEATURE_AVX2

def supports(uint32_t features):
    return CpuInfo.supports(features)



from __future__ import annotations

from rapidfuzz import utils_cpp, utils_py
from tests.distance.common import Levenshtein


class CustomHashable:
    def __init__(self, string):
        self._string = string

    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._string)


def test_empty_string():
    """
    when both strings are empty this is a perfect match
    """
    assert Levenshtein.distance("", "") == 0
    assert Levenshtein.distance("", "", weights=(1, 1, 0)) == 0
    assert Levenshtein.distance("", "", weights=(1, 1, 2)) == 0
    assert Levenshtein.distance("", "", weights=(1, 1, 5)) == 0
    assert Levenshtein.distance("", "", weights=(3, 7, 5)) == 0


def test_score_cutoff_overflow():
    """
    when score_cutoff was out of bounds for size_t we had an overflow error
    """
    assert Levenshtein.distance("", "") == 0
    assert Levenshtein.distance("", "", score_cutoff=2**63) == 0


def test_cross_type_matching():
    """
    strings should always be interpreted in the same way
    """
    assert Levenshtein.distance("aaaa", "aaaa") == 0
    assert Levenshtein.distance("aaaa", ["a", "a", "a", "a"]) == 0
    assert Levenshtein.distance(b"aaaa", b"aaaa") == 0
    assert Levenshtein.distance("aaaa", b"aaaa") == 0
    # todo add support in pure python
    assert Levenshtein.distance("aaaa", [ord("a"), ord("a"), "a", "a"]) == 0
    assert Levenshtein.distance([0, -1], [0, -2]) == 1
    assert (
        Levenshtein.distance(
            [CustomHashable("aa"), CustomHashable("aa")],
            [CustomHashable("aa"), CustomHashable("bb")],
        )
        == 1
    )


def test_word_error_rate():
    """
    it should be possible to use levenshtein to implement a word error rate
    """
    assert Levenshtein.distance(["aaaaa", "bbbb"], ["aaaaa", "bbbb"]) == 0
    assert Levenshtein.distance(["aaaaa", "bbbb"], ["aaaaa", "cccc"]) == 1


def test_simple_unicode_tests():
    """
    some very simple tests using unicode with scorers
    to catch relatively obvious implementation errors
    """
    s1 = "ÁÄ"
    s2 = "ABCD"
    assert Levenshtein.distance(s1, s2) == 4  # 2 sub + 2 ins
    assert Levenshtein.distance(s1, s2, weights=(1, 1, 0)) == 2  # 2 sub + 2 ins
    assert Levenshtein.distance(s1, s2, weights=(1, 1, 2)) == 6  # 2 del + 4 ins / 2 sub + 2 ins
    assert Levenshtein.distance(s1, s2, weights=(1, 1, 5)) == 6  # 2 del + 4 ins
    assert Levenshtein.distance(s1, s2, weights=(1, 7, 5)) == 12  # 2 sub + 2 ins
    assert Levenshtein.distance(s2, s1, weights=(1, 7, 5)) == 24  # 2 sub + 2 del

    assert Levenshtein.distance(s1, s1) == 0
    assert Levenshtein.distance(s1, s1, weights=(1, 1, 0)) == 0
    assert Levenshtein.distance(s1, s1, weights=(1, 1, 2)) == 0
    assert Levenshtein.distance(s1, s1, weights=(1, 1, 5)) == 0
    assert Levenshtein.distance(s1, s1, weights=(3, 7, 5)) == 0


def test_Editops():
    """
    basic test for Levenshtein.editops
    """
    assert Levenshtein.editops("0", "").as_list() == [("delete", 0, 0)]
    assert Levenshtein.editops("", "0").as_list() == [("insert", 0, 0)]

    assert Levenshtein.editops("00", "0").as_list() == [("delete", 1, 1)]
    assert Levenshtein.editops("0", "00").as_list() == [("insert", 1, 1)]

    assert Levenshtein.editops("qabxcd", "abycdf").as_list() == [
        ("delete", 0, 0),
        ("replace", 3, 2),
        ("insert", 6, 5),
    ]
    assert Levenshtein.editops("Lorem ipsum.", "XYZLorem ABC iPsum").as_list() == [
        ("insert", 0, 0),
        ("insert", 0, 1),
        ("insert", 0, 2),
        ("insert", 6, 9),
        ("insert", 6, 10),
        ("insert", 6, 11),
        ("insert", 6, 12),
        ("replace", 7, 14),
        ("delete", 11, 18),
    ]

    ops = Levenshtein.editops("aaabaaa", "abbaaabba")
    assert ops.src_len == 7
    assert ops.dest_len == 9


def test_Opcodes():
    """
    basic test for Levenshtein.opcodes
    """
    assert Levenshtein.opcodes("", "abc").as_list() == [("insert", 0, 0, 0, 3)]


def test_mbleven():
    """
    test for regressions to previous bugs in the cached Levenshtein implementation
    """
    assert Levenshtein.distance("0", "101", score_cutoff=1) == 2
    assert Levenshtein.distance("0", "101", score_cutoff=2) == 2
    assert Levenshtein.distance("0", "101", score_cutoff=3) == 2


def testCaseInsensitive():
    assert Levenshtein.distance("new york mets", "new YORK mets", processor=utils_cpp.default_process) == 0
    assert Levenshtein.distance("new york mets", "new YORK mets", processor=utils_py.default_process) == 0



from __future__ import annotations

import subprocess

from PyInstaller import __main__ as pyi_main


# Test out the package by importing it, then running functions from it.
def test_pyi_hooksample(tmp_path):
    app_name = "userapp"
    workpath = tmp_path / "build"
    distpath = tmp_path / "dist"
    app = tmp_path / (app_name + ".py")
    app.write_text(
        "\n".join(
            [
                "import rapidfuzz",
                "from rapidfuzz.distance import metrics_py",
                "from rapidfuzz.distance import metrics_cpp",
                "rapidfuzz.distance.Levenshtein.distance('test', 'teste')",
                "metrics_py.levenshtein_distance('test', 'teste')",
                "metrics_cpp.levenshtein_distance('test', 'teste')",
            ]
        )
    )
    args = [
        # Place all generated files in ``tmp_path``.
        "--workpath",
        str(workpath),
        "--distpath",
        str(distpath),
        "--specpath",
        str(tmp_path),
        str(app),
    ]
    pyi_main.run(args)
    subprocess.run([str(distpath / app_name / app_name)], check=True)



# SPDX-License-Identifier: MIT
# Copyright (C) 2022 Max Bachmann
from __future__ import annotations

__all__ = ["AVX2", "SSE2", "supports"]

try:
    from rapidfuzz._feature_detector_cpp import AVX2, SSE2, supports
except ImportError:
    SSE2 = 1
    AVX2 = 2

    def supports(features):
        _ = features
        return False



# SPDX-License-Identifier: MIT
# Copyright (C) 2022 Max Bachmann
from __future__ import annotations

from math import ceil

from rapidfuzz._common_py import conv_sequences
from rapidfuzz._utils import ScorerFlag, add_scorer_attrs, is_none, setupPandas
from rapidfuzz.distance import ScoreAlignment
from rapidfuzz.distance.Indel_py import (
    _block_normalized_similarity as indel_block_normalized_similarity,
    distance as indel_distance,
    normalized_similarity as indel_normalized_similarity,
)


def get_scorer_flags_fuzz(**_kwargs):
    return {
        "optimal_score": 100,
        "worst_score": 0,
        "flags": ScorerFlag.RESULT_F64 | ScorerFlag.SYMMETRIC,
    }


fuzz_attribute = {"get_scorer_flags": get_scorer_flags_fuzz}


def _norm_distance(dist, lensum, score_cutoff):
    score = (100 - 100 * dist / lensum) if lensum else 100
    return score if score >= score_cutoff else 0


def _split_sequence(seq):
    if isinstance(seq, (str, bytes)):
        return seq.split()

    splitted_seq = [[]]
    for x in seq:
        ch = x if isinstance(x, str) else chr(x)
        if ch.isspace():
            splitted_seq.append([])
        else:
            splitted_seq[-1].append(x)

    return [tuple(x) for x in splitted_seq if x]


def _join_splitted_sequence(seq_list):
    if not seq_list:
        return ""
    if isinstance(next(iter(seq_list)), str):
        return " ".join(seq_list)
    if isinstance(next(iter(seq_list)), bytes):
        return b" ".join(seq_list)

    joined = []
    for seq in seq_list:
        joined += seq
        joined += [ord(" ")]
    return joined[:-1]


def ratio(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Calculates the normalized Indel similarity.

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 100.
        For ratio < score_cutoff 0 is returned instead. Default is 0,
        which deactivates this behaviour.

    Returns
    -------
    similarity : float
        similarity between s1 and s2 as a float between 0 and 100

    See Also
    --------
    rapidfuzz.distance.Indel.normalized_similarity : Normalized Indel similarity

    Notes
    -----
    .. image:: img/ratio.svg

    Examples
    --------
    >>> fuzz.ratio("this is a test", "this is a test!")
    96.55171966552734
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0

    if score_cutoff is not None:
        score_cutoff /= 100

    score = indel_normalized_similarity(s1, s2, processor=processor, score_cutoff=score_cutoff)
    return score * 100


def _partial_ratio_impl(s1, s2, score_cutoff):
    """
    implementation of partial_ratio. This assumes len(s1) <= len(s2).
    """
    s1_char_set = set(s1)
    len1 = len(s1)
    len2 = len(s2)

    res = ScoreAlignment(0, 0, len1, 0, len1)

    block = {}
    block_get = block.get
    x = 1
    for ch1 in s1:
        block[ch1] = block_get(ch1, 0) | x
        x <<= 1

    for i in range(1, len1):
        substr_last = s2[i - 1]
        if substr_last not in s1_char_set:
            continue

        # todo cache map
        ls_ratio = indel_block_normalized_similarity(block, s1, s2[:i], score_cutoff=score_cutoff)
        if ls_ratio > res.score:
            res.score = score_cutoff = ls_ratio
            res.dest_start = 0
            res.dest_end = i
            if res.score == 1:
                res.score = 100
                return res

    for i in range(len2 - len1):
        substr_last = s2[i + len1 - 1]
        if substr_last not in s1_char_set:
            continue

        # todo cache map
        ls_ratio = indel_block_normalized_similarity(block, s1, s2[i : i + len1], score_cutoff=score_cutoff)
        if ls_ratio > res.score:
            res.score = score_cutoff = ls_ratio
            res.dest_start = i
            res.dest_end = i + len1
            if res.score == 1:
                res.score = 100
                return res

    for i in range(len2 - len1, len2):
        substr_first = s2[i]
        if substr_first not in s1_char_set:
            continue

        # todo cache map
        ls_ratio = indel_block_normalized_similarity(block, s1, s2[i:], score_cutoff=score_cutoff)
        if ls_ratio > res.score:
            res.score = score_cutoff = ls_ratio
            res.dest_start = i
            res.dest_end = len2
            if res.score == 1:
                res.score = 100
                return res

    res.score *= 100
    return res


def partial_ratio(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Searches for the optimal alignment of the shorter string in the
    longer string and returns the fuzz.ratio for this alignment.

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 100.
        For ratio < score_cutoff 0 is returned instead. Default is 0,
        which deactivates this behaviour.

    Returns
    -------
    similarity : float
        similarity between s1 and s2 as a float between 0 and 100

    Notes
    -----
    Depending on the length of the needle (shorter string) different
    implementations are used to improve the performance.

    short needle (length ≤ 64):
        When using a short needle length the fuzz.ratio is calculated for all
        alignments that could result in an optimal alignment. It is
        guaranteed to find the optimal alignment. For short needles this is very
        fast, since for them fuzz.ratio runs in ``O(N)`` time. This results in a worst
        case performance of ``O(NM)``.

    .. image:: img/partial_ratio_short_needle.svg

    long needle (length > 64):
        For long needles a similar implementation to FuzzyWuzzy is used.
        This implementation only considers alignments which start at one
        of the longest common substrings. This results in a worst case performance
        of ``O(N[N/64]M)``. However usually most of the alignments can be skipped.
        The following Python code shows the concept:

        .. code-block:: python

            blocks = SequenceMatcher(None, needle, longer, False).get_matching_blocks()
            score = 0
            for block in blocks:
                long_start = block[1] - block[0] if (block[1] - block[0]) > 0 else 0
                long_end = long_start + len(shorter)
                long_substr = longer[long_start:long_end]
                score = max(score, fuzz.ratio(needle, long_substr))

        This is a lot faster than checking all possible alignments. However it
        only finds one of the best alignments and not necessarily the optimal one.

    .. image:: img/partial_ratio_long_needle.svg

    Examples
    --------
    >>> fuzz.partial_ratio("this is a test", "this is a test!")
    100.0
    """
    alignment = partial_ratio_alignment(s1, s2, processor=processor, score_cutoff=score_cutoff)
    if alignment is None:
        return 0

    return alignment.score


def partial_ratio_alignment(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Searches for the optimal alignment of the shorter string in the
    longer string and returns the fuzz.ratio and the corresponding
    alignment.

    Parameters
    ----------
    s1 : str | bytes
        First string to compare.
    s2 : str | bytes
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 100.
        For ratio < score_cutoff None is returned instead. Default is 0,
        which deactivates this behaviour.

    Returns
    -------
    alignment : ScoreAlignment, optional
        alignment between s1 and s2 with the score as a float between 0 and 100

    Examples
    --------
    >>> s1 = "a certain string"
    >>> s2 = "cetain"
    >>> res = fuzz.partial_ratio_alignment(s1, s2)
    >>> res
    ScoreAlignment(score=83.33333333333334, src_start=2, src_end=8, dest_start=0, dest_end=6)

    Using the alignment information it is possible to calculate the same fuzz.ratio

    >>> fuzz.ratio(s1[res.src_start:res.src_end], s2[res.dest_start:res.dest_end])
    83.33333333333334
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return None

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    if score_cutoff is None:
        score_cutoff = 0

    if not s1 and not s2:
        return ScoreAlignment(100.0, 0, 0, 0, 0)
    s1, s2 = conv_sequences(s1, s2)
    len1 = len(s1)
    len2 = len(s2)
    if len1 <= len2:
        shorter = s1
        longer = s2
    else:
        shorter = s2
        longer = s1

    res = _partial_ratio_impl(shorter, longer, score_cutoff / 100)
    if res.score != 100 and len1 == len2:
        score_cutoff = max(score_cutoff, res.score)
        res2 = _partial_ratio_impl(longer, shorter, score_cutoff / 100)
        if res2.score > res.score:
            res = ScoreAlignment(res2.score, res2.dest_start, res2.dest_end, res2.src_start, res2.src_end)

    if res.score < score_cutoff:
        return None

    if len1 <= len2:
        return res

    return ScoreAlignment(res.score, res.dest_start, res.dest_end, res.src_start, res.src_end)


def token_sort_ratio(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Sorts the words in the strings and calculates the fuzz.ratio between them

    Parameters
    ----------
    s1 : str
        First string to compare.
    s2 : str
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 100.
        For ratio < score_cutoff 0 is returned instead. Default is 0,
        which deactivates this behaviour.

    Returns
    -------
    similarity : float
        similarity between s1 and s2 as a float between 0 and 100

    Notes
    -----
    .. image:: img/token_sort_ratio.svg

    Examples
    --------
    >>> fuzz.token_sort_ratio("fuzzy wuzzy was a bear", "wuzzy fuzzy was a bear")
    100.0
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    s1, s2 = conv_sequences(s1, s2)
    sorted_s1 = _join_splitted_sequence(sorted(_split_sequence(s1)))
    sorted_s2 = _join_splitted_sequence(sorted(_split_sequence(s2)))
    return ratio(sorted_s1, sorted_s2, score_cutoff=score_cutoff)


def token_set_ratio(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Compares the words in the strings based on unique and common words between them
    using fuzz.ratio

    Parameters
    ----------
    s1 : str
        First string to compare.
    s2 : str
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 100.
        For ratio < score_cutoff 0 is returned instead. Default is 0,
        which deactivates this behaviour.

    Returns
    -------
    similarity : float
        similarity between s1 and s2 as a float between 0 and 100

    Notes
    -----
    .. image:: img/token_set_ratio.svg

    Examples
    --------
    >>> fuzz.token_sort_ratio("fuzzy was a bear", "fuzzy fuzzy was a bear")
    83.8709716796875
    >>> fuzz.token_set_ratio("fuzzy was a bear", "fuzzy fuzzy was a bear")
    100.0
    # Returns 100.0 if one string is a subset of the other, regardless of extra content in the longer string
    >>> fuzz.token_set_ratio("fuzzy was a bear but not a dog", "fuzzy was a bear")
    100.0
    # Score is reduced only when there is explicit disagreement in the two strings
    >>> fuzz.token_set_ratio("fuzzy was a bear but not a dog", "fuzzy was a bear but not a cat")
    92.3076923076923
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    if score_cutoff is None:
        score_cutoff = 0

    s1, s2 = conv_sequences(s1, s2)

    tokens_a = set(_split_sequence(s1))
    tokens_b = set(_split_sequence(s2))

    # in FuzzyWuzzy this returns 0. For sake of compatibility return 0 here as well
    # see https://github.com/rapidfuzz/RapidFuzz/issues/110
    if not tokens_a or not tokens_b:
        return 0

    intersect = tokens_a.intersection(tokens_b)
    diff_ab = tokens_a.difference(tokens_b)
    diff_ba = tokens_b.difference(tokens_a)

    # one sentence is part of the other one
    if intersect and (not diff_ab or not diff_ba):
        return 100

    diff_ab_joined = _join_splitted_sequence(sorted(diff_ab))
    diff_ba_joined = _join_splitted_sequence(sorted(diff_ba))

    ab_len = len(diff_ab_joined)
    ba_len = len(diff_ba_joined)
    # todo is length sum without joining faster?
    sect_len = len(_join_splitted_sequence(intersect))

    # string length sect+ab <-> sect and sect+ba <-> sect
    sect_ab_len = sect_len + (sect_len != 0) + ab_len
    sect_ba_len = sect_len + (sect_len != 0) + ba_len

    result = 0.0
    cutoff_distance = ceil((sect_ab_len + sect_ba_len) * (1 - score_cutoff / 100))
    dist = indel_distance(diff_ab_joined, diff_ba_joined, score_cutoff=cutoff_distance)

    if dist <= cutoff_distance:
        result = _norm_distance(dist, sect_ab_len + sect_ba_len, score_cutoff)

    # exit early since the other ratios are 0
    if not sect_len:
        return result

    # levenshtein distance sect+ab <-> sect and sect+ba <-> sect
    # since only sect is similar in them the distance can be calculated based on
    # the length difference
    sect_ab_dist = (sect_len != 0) + ab_len
    sect_ab_ratio = _norm_distance(sect_ab_dist, sect_len + sect_ab_len, score_cutoff)

    sect_ba_dist = (sect_len != 0) + ba_len
    sect_ba_ratio = _norm_distance(sect_ba_dist, sect_len + sect_ba_len, score_cutoff)

    return max(result, sect_ab_ratio, sect_ba_ratio)


def token_ratio(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Helper method that returns the maximum of fuzz.token_set_ratio and fuzz.token_sort_ratio
    (faster than manually executing the two functions)

    Parameters
    ----------
    s1 : str
        First string to compare.
    s2 : str
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 100.
        For ratio < score_cutoff 0 is returned instead. Default is 0,
        which deactivates this behaviour.

    Returns
    -------
    similarity : float
        similarity between s1 and s2 as a float between 0 and 100

    Notes
    -----
    .. image:: img/token_ratio.svg
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    # todo write combined implementation
    return max(
        token_set_ratio(s1, s2, processor=None, score_cutoff=score_cutoff),
        token_sort_ratio(s1, s2, processor=None, score_cutoff=score_cutoff),
    )


def partial_token_sort_ratio(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    sorts the words in the strings and calculates the fuzz.partial_ratio between them

    Parameters
    ----------
    s1 : str
        First string to compare.
    s2 : str
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 100.
        For ratio < score_cutoff 0 is returned instead. Default is 0,
        which deactivates this behaviour.

    Returns
    -------
    similarity : float
        similarity between s1 and s2 as a float between 0 and 100

    Notes
    -----
    .. image:: img/partial_token_sort_ratio.svg
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    s1, s2 = conv_sequences(s1, s2)
    sorted_s1 = _join_splitted_sequence(sorted(_split_sequence(s1)))
    sorted_s2 = _join_splitted_sequence(sorted(_split_sequence(s2)))
    return partial_ratio(sorted_s1, sorted_s2, score_cutoff=score_cutoff)


def partial_token_set_ratio(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Compares the words in the strings based on unique and common words between them
    using fuzz.partial_ratio

    Parameters
    ----------
    s1 : str
        First string to compare.
    s2 : str
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 100.
        For ratio < score_cutoff 0 is returned instead. Default is 0,
        which deactivates this behaviour.

    Returns
    -------
    similarity : float
        similarity between s1 and s2 as a float between 0 and 100

    Notes
    -----
    .. image:: img/partial_token_set_ratio.svg
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    s1, s2 = conv_sequences(s1, s2)

    tokens_a = set(_split_sequence(s1))
    tokens_b = set(_split_sequence(s2))
    # in FuzzyWuzzy this returns 0. For sake of compatibility return 0 here as well
    # see https://github.com/rapidfuzz/RapidFuzz/issues/110
    if not tokens_a or not tokens_b:
        return 0

    # exit early when there is a common word in both sequences
    if tokens_a.intersection(tokens_b):
        return 100

    diff_ab = _join_splitted_sequence(sorted(tokens_a.difference(tokens_b)))
    diff_ba = _join_splitted_sequence(sorted(tokens_b.difference(tokens_a)))
    return partial_ratio(diff_ab, diff_ba, score_cutoff=score_cutoff)


def partial_token_ratio(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Helper method that returns the maximum of fuzz.partial_token_set_ratio and
    fuzz.partial_token_sort_ratio (faster than manually executing the two functions)

    Parameters
    ----------
    s1 : str
        First string to compare.
    s2 : str
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 100.
        For ratio < score_cutoff 0 is returned instead. Default is 0,
        which deactivates this behaviour.

    Returns
    -------
    similarity : float
        similarity between s1 and s2 as a float between 0 and 100

    Notes
    -----
    .. image:: img/partial_token_ratio.svg
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    if score_cutoff is None:
        score_cutoff = 0

    s1, s2 = conv_sequences(s1, s2)

    tokens_split_a = _split_sequence(s1)
    tokens_split_b = _split_sequence(s2)
    tokens_a = set(tokens_split_a)
    tokens_b = set(tokens_split_b)

    # exit early when there is a common word in both sequences
    if tokens_a.intersection(tokens_b):
        return 100

    diff_ab = tokens_a.difference(tokens_b)
    diff_ba = tokens_b.difference(tokens_a)

    result = partial_ratio(
        _join_splitted_sequence(sorted(tokens_split_a)),
        _join_splitted_sequence(sorted(tokens_split_b)),
        score_cutoff=score_cutoff,
    )

    # do not calculate the same partial_ratio twice
    if len(tokens_split_a) == len(diff_ab) and len(tokens_split_b) == len(diff_ba):
        return result

    score_cutoff = max(score_cutoff, result)
    return max(
        result,
        partial_ratio(
            _join_splitted_sequence(sorted(diff_ab)),
            _join_splitted_sequence(sorted(diff_ba)),
            score_cutoff=score_cutoff,
        ),
    )


def WRatio(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Calculates a weighted ratio based on the other ratio algorithms

    Parameters
    ----------
    s1 : str
        First string to compare.
    s2 : str
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 100.
        For ratio < score_cutoff 0 is returned instead. Default is 0,
        which deactivates this behaviour.

    Returns
    -------
    similarity : float
        similarity between s1 and s2 as a float between 0 and 100

    Notes
    -----
    .. image:: img/WRatio.svg
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0

    UNBASE_SCALE = 0.95

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    # in FuzzyWuzzy this returns 0. For sake of compatibility return 0 here as well
    # see https://github.com/rapidfuzz/RapidFuzz/issues/110
    if not s1 or not s2:
        return 0

    if score_cutoff is None:
        score_cutoff = 0

    len1 = len(s1)
    len2 = len(s2)
    len_ratio = len1 / len2 if len1 > len2 else len2 / len1

    end_ratio = ratio(s1, s2, score_cutoff=score_cutoff)
    if len_ratio < 1.5:
        score_cutoff = max(score_cutoff, end_ratio) / UNBASE_SCALE
        return max(
            end_ratio,
            token_ratio(s1, s2, score_cutoff=score_cutoff, processor=None) * UNBASE_SCALE,
        )

    PARTIAL_SCALE = 0.9 if len_ratio < 8.0 else 0.6
    score_cutoff = max(score_cutoff, end_ratio) / PARTIAL_SCALE
    end_ratio = max(end_ratio, partial_ratio(s1, s2, score_cutoff=score_cutoff) * PARTIAL_SCALE)

    score_cutoff = max(score_cutoff, end_ratio) / UNBASE_SCALE
    return max(
        end_ratio,
        partial_token_ratio(s1, s2, score_cutoff=score_cutoff, processor=None) * UNBASE_SCALE * PARTIAL_SCALE,
    )


def QRatio(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Calculates a quick ratio between two strings using fuzz.ratio.

    Since v3.0 this behaves similar to fuzz.ratio with the exception that this
    returns 0 when comparing two empty strings

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 100.
        For ratio < score_cutoff 0 is returned instead. Default is 0,
        which deactivates this behaviour.

    Returns
    -------
    similarity : float
        similarity between s1 and s2 as a float between 0 and 100

    Examples
    --------
    >>> fuzz.QRatio("this is a test", "this is a test!")
    96.55171966552734
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)
    # in FuzzyWuzzy this returns 0. For sake of compatibility return 0 here as well
    # see https://github.com/rapidfuzz/RapidFuzz/issues/110
    if not s1 or not s2:
        return 0

    return ratio(s1, s2, score_cutoff=score_cutoff)


add_scorer_attrs(ratio, fuzz_attribute)
add_scorer_attrs(partial_ratio, fuzz_attribute)
add_scorer_attrs(token_sort_ratio, fuzz_attribute)
add_scorer_attrs(token_set_ratio, fuzz_attribute)
add_scorer_attrs(token_ratio, fuzz_attribute)
add_scorer_attrs(partial_token_sort_ratio, fuzz_attribute)
add_scorer_attrs(partial_token_set_ratio, fuzz_attribute)
add_scorer_attrs(partial_token_ratio, fuzz_attribute)
add_scorer_attrs(WRatio, fuzz_attribute)
add_scorer_attrs(QRatio, fuzz_attribute)



# SPDX-License-Identifier: MIT
# Copyright (C) 2022 Max Bachmann

from __future__ import annotations

import sys
from math import isnan
from typing import Any, Callable

pandas_NA = None


def setupPandas():
    global pandas_NA  # noqa: PLW0603
    if pandas_NA is None:
        pandas = sys.modules.get("pandas")
        if hasattr(pandas, "NA"):
            pandas_NA = pandas.NA


setupPandas()


class ScorerFlag:
    RESULT_F64 = 1 << 5
    RESULT_I64 = 1 << 6
    RESULT_SIZE_T = 1 << 7
    SYMMETRIC = 1 << 11


def _get_scorer_flags_distance(**_kwargs: Any) -> dict[str, Any]:
    return {
        "optimal_score": 0,
        "worst_score": 2**63 - 1,
        "flags": ScorerFlag.RESULT_SIZE_T | ScorerFlag.SYMMETRIC,
    }


def _get_scorer_flags_similarity(**_kwargs: Any) -> dict[str, Any]:
    return {
        "optimal_score": 2**63 - 1,
        "worst_score": 0,
        "flags": ScorerFlag.RESULT_SIZE_T | ScorerFlag.SYMMETRIC,
    }


def _get_scorer_flags_normalized_distance(**_kwargs: Any) -> dict[str, Any]:
    return {
        "optimal_score": 0,
        "worst_score": 1,
        "flags": ScorerFlag.RESULT_F64 | ScorerFlag.SYMMETRIC,
    }


def _get_scorer_flags_normalized_similarity(**_kwargs: Any) -> dict[str, Any]:
    return {
        "optimal_score": 1,
        "worst_score": 0,
        "flags": ScorerFlag.RESULT_F64 | ScorerFlag.SYMMETRIC,
    }


def is_none(s: Any) -> bool:
    if s is None or s is pandas_NA:
        return True

    if isinstance(s, float) and isnan(s):
        return True

    return False


def add_scorer_attrs(func: Any, cached_scorer_call: dict[str, Callable[..., dict[str, Any]]]):
    func._RF_ScorerPy = cached_scorer_call
    # used to detect the function hasn't been wrapped afterwards
    func._RF_OriginalScorer = func


default_distance_attribute: dict[str, Callable[..., dict[str, Any]]] = {"get_scorer_flags": _get_scorer_flags_distance}
default_similarity_attribute: dict[str, Callable[..., dict[str, Any]]] = {
    "get_scorer_flags": _get_scorer_flags_similarity
}
default_normalized_distance_attribute: dict[str, Callable[..., dict[str, Any]]] = {
    "get_scorer_flags": _get_scorer_flags_normalized_distance
}
default_normalized_similarity_attribute: dict[str, Callable[..., dict[str, Any]]] = {
    "get_scorer_flags": _get_scorer_flags_normalized_similarity
}



from __future__ import annotations

from array import array

import pytest

from tests.distance.common import all_scorer_modules


class CustomHashable:
    def __init__(self, string):
        self._string = string

    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._string)


@pytest.mark.parametrize("scorer", all_scorer_modules)
def test_none(scorer):
    """
    All normalized scorers should be able to handle None values
    """
    assert scorer.normalized_distance(None, "test") == 1.0
    assert scorer.normalized_similarity(None, "test") == 0.0
    assert scorer.normalized_distance("test", None) == 1.0
    assert scorer.normalized_similarity("test", None) == 0.0


@pytest.mark.parametrize("scorer", all_scorer_modules)
def test_nan(scorer):
    """
    All normalized scorers should be able to handle float("nan")
    """
    assert scorer.normalized_distance(float("nan"), "test") == 1.0
    assert scorer.normalized_similarity(float("nan"), "test") == 0.0
    assert scorer.normalized_distance("test", float("nan")) == 1.0
    assert scorer.normalized_similarity("test", float("nan")) == 0.0


@pytest.mark.parametrize("scorer", all_scorer_modules)
def test_empty_strings(scorer):
    """
    Test behavior when comparing two empty strings
    """
    assert scorer.normalized_distance("", "") == 0.0
    assert scorer.normalized_similarity("", "") == 1.0


@pytest.mark.parametrize("scorer", all_scorer_modules)
def test_similar_array(scorer):
    """
    arrays should be supported and treated in a compatible way to strings
    """
    assert (
        scorer.normalized_similarity(
            array("u", "the wonderful new york mets"),
            array("u", "the wonderful new york mets"),
        )
        == 1.0
    )
    assert scorer.normalized_similarity("the wonderful new york mets", array("u", "the wonderful new york mets")) == 1.0
    assert scorer.normalized_similarity(array("u", "the wonderful new york mets"), "the wonderful new york mets") == 1.0


@pytest.mark.parametrize("scorer", all_scorer_modules)
def test_similar_bytes(scorer):
    """
    bytes should be supported and treated in a compatible way to strings
    """
    assert scorer.normalized_similarity(b"the wonderful new york mets", b"the wonderful new york mets") == 1.0
    assert scorer.normalized_similarity("the wonderful new york mets", b"the wonderful new york mets") == 1.0
    assert scorer.normalized_similarity(b"the wonderful new york mets", "the wonderful new york mets") == 1.0


@pytest.mark.parametrize("scorer", all_scorer_modules)
def test_similar_ord_array(scorer):
    """
    elements in string should behave similar to their underlying unicode number
    """
    assert (
        scorer.normalized_similarity(
            [ord("a"), ord("a"), "a", "a"],
            [ord("a"), ord("a"), "a", "a"],
        )
        == 1.0
    )
    assert scorer.normalized_similarity("aaaa", [ord("a"), ord("a"), "a", "a"]) == 1.0
    assert scorer.normalized_similarity([ord("a"), ord("a"), "a", "a"], "aaaa") == 1.0


@pytest.mark.parametrize("scorer", all_scorer_modules)
def test_integer_array(scorer):
    """
    the numbers -1 and -2 should be treated differently (these are hash collisions in the standard has)
    """
    assert scorer.normalized_similarity([0, -1], [0, -2]) != 1.0


@pytest.mark.parametrize("scorer", all_scorer_modules)
def test_custom_hashable(scorer):
    """
    custom types should be compared using their hash
    """
    assert (
        scorer.normalized_similarity(
            [CustomHashable("aa"), CustomHashable("aa")],
            [CustomHashable("aa"), CustomHashable("aa")],
        )
        == 1.0
    )

    assert (
        scorer.normalized_similarity(
            [CustomHashable("aa"), CustomHashable("aa")],
            [CustomHashable("aa"), CustomHashable("bb")],
        )
        != 1.0
    )



# SPDX-License-Identifier: MIT
# Copyright (C) 2022 Max Bachmann
from __future__ import annotations

from rapidfuzz._common_py import conv_sequences
from rapidfuzz._utils import is_none, setupPandas


def _jaro_calculate_similarity(pattern_len, text_len, common_chars, transpositions):
    transpositions //= 2
    sim = 0.0
    sim += common_chars / pattern_len
    sim += common_chars / text_len
    sim += (common_chars - transpositions) / common_chars
    return sim / 3.0


def _jaro_length_filter(pattern_len, text_len, score_cutoff):
    """
    filter matches below score_cutoff based on string lengths
    """
    if not pattern_len or not text_len:
        return False

    sim = _jaro_calculate_similarity(pattern_len, text_len, min(pattern_len, text_len), 0)
    return sim >= score_cutoff


def _jaro_common_char_filter(pattern_len, text_len, common_chars, score_cutoff):
    """
    filter matches below score_cutoff based on string lengths and common characters
    """
    if not common_chars:
        return False

    sim = _jaro_calculate_similarity(pattern_len, text_len, common_chars, 0)
    return sim >= score_cutoff


def _jaro_bounds(s1, s2):
    """
    find bounds and skip out of bound parts of the sequences
    """
    pattern_len = len(s1)
    text_len = len(s2)

    # since jaro uses a sliding window some parts of T/P might never be in
    # range an can be removed ahead of time
    bound = 0
    if text_len > pattern_len:
        bound = text_len // 2 - 1
        if text_len > pattern_len + bound:
            s2 = s2[: pattern_len + bound]
    else:
        bound = pattern_len // 2 - 1
        if pattern_len > text_len + bound:
            s1 = s1[: text_len + bound]
    return s1, s2, bound


def similarity(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Calculates the jaro similarity

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 1.0.
        For ratio < score_cutoff 0 is returned instead. Default is None,
        which deactivates this behaviour.

    Returns
    -------
    similarity : float
        similarity between s1 and s2 as a float between 0 and 1.0
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    if not s1 and not s2:
        return 1.0

    if score_cutoff is None:
        score_cutoff = 0

    s1, s2 = conv_sequences(s1, s2)
    pattern_len = len(s1)
    text_len = len(s2)

    # short circuit if score_cutoff can not be reached
    if not _jaro_length_filter(pattern_len, text_len, score_cutoff):
        return 0

    if pattern_len == 1 and text_len == 1:
        return float(s1[0] == s2[0])

    s1, s2, bound = _jaro_bounds(s1, s2)

    s1_flags = [False] * pattern_len
    s2_flags = [False] * text_len

    # todo use bitparallel implementation
    # looking only within search range, count & flag matched pairs
    common_chars = 0
    for i, s1_ch in enumerate(s1):
        low = max(0, i - bound)
        hi = min(i + bound, text_len - 1)
        for j in range(low, hi + 1):
            if not s2_flags[j] and s2[j] == s1_ch:
                s1_flags[i] = s2_flags[j] = True
                common_chars += 1
                break

    # short circuit if score_cutoff can not be reached
    if not _jaro_common_char_filter(pattern_len, text_len, common_chars, score_cutoff):
        return 0

    # todo use bitparallel implementation
    # count transpositions
    k = trans_count = 0
    for i, s1_f in enumerate(s1_flags):
        if s1_f:
            for j in range(k, text_len):
                if s2_flags[j]:
                    k = j + 1
                    break
            if s1[i] != s2[j]:
                trans_count += 1

    return _jaro_calculate_similarity(pattern_len, text_len, common_chars, trans_count)


def normalized_similarity(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Calculates the normalized jaro similarity

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 1.0.
        For ratio < score_cutoff 0 is returned instead. Default is None,
        which deactivates this behaviour.

    Returns
    -------
    normalized similarity : float
        normalized similarity between s1 and s2 as a float between 0 and 1.0
    """
    return similarity(s1, s2, processor=processor, score_cutoff=score_cutoff)


def distance(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Calculates the jaro distance

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 1.0.
        For ratio < score_cutoff 0 is returned instead. Default is None,
        which deactivates this behaviour.

    Returns
    -------
    distance : float
        distance between s1 and s2 as a float between 1.0 and 0.0
    """
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    cutoff_distance = None if (score_cutoff is None or score_cutoff > 1.0) else 1.0 - score_cutoff
    sim = similarity(s1, s2, score_cutoff=cutoff_distance)
    dist = 1.0 - sim
    return dist if (score_cutoff is None or dist <= score_cutoff) else 1.0


def normalized_distance(
    s1,
    s2,
    *,
    processor=None,
    score_cutoff=None,
):
    """
    Calculates the normalized jaro distance

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    processor: callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 1.0.
        For ratio < score_cutoff 0 is returned instead. Default is None,
        which deactivates this behaviour.

    Returns
    -------
    normalized distance : float
        normalized distance between s1 and s2 as a float between 1.0 and 0.0
    """
    return distance(s1, s2, processor=processor, score_cutoff=score_cutoff)



from __future__ import annotations

import importlib
import random
import string
from timeit import timeit

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

random.seed(18)

plt.rc("font", size=13)  # controls default text sizes
plt.rc("axes", titlesize=18)  # fontsize of the axes title
plt.rc("axes", labelsize=15)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=15)  # fontsize of the tick labels
plt.rc("ytick", labelsize=15)  # fontsize of the tick labels
plt.rc("legend", fontsize=15)  # legend fontsize

PROCESSOR = {
    "ratio": False,
    "partial_ratio": False,
    "token_sort_ratio": True,
    "token_set_ratio": True,
    "partial_token_sort_ratio": True,
    "partial_token_set_ratio": True,
    "QRatio": True,
    "WRatio": True,
}

LIBRARIES = (
    "ratio",
    "partial_ratio",
    "token_sort_ratio",
    "token_set_ratio",
    "partial_token_sort_ratio",
    "partial_token_set_ratio",
    "QRatio",
    "WRatio",
)


def load_func(target):
    modname, funcname = target.rsplit(".", maxsplit=1)

    module = importlib.import_module(modname)
    return getattr(module, funcname)


def get_platform():
    import platform

    uname = platform.uname()
    pyver = platform.python_version()
    return f"Python {pyver} on {uname.system} ({uname.machine})"


def benchmark():
    words = ["".join(random.choice(string.ascii_letters + string.digits) for _ in range(8)) for _ in range(10000)]
    sample_rate = len(words) // 100
    sample = words[::sample_rate]
    total = len(words) * len(sample)

    print("System:", get_platform())
    print("Words :", len(words))
    print("Sample:", len(sample))
    print("Total : %s calls\n" % total)

    def wrap_cdist(scorer, processor):
        from rapidfuzz.process import cdist

        def func():
            cdist(sample, words, scorer=scorer, processor=processor)

        return func

    def wrap_iterate(scorer, processor):
        if processor:
            queries = [processor(x) for x in sample]
            choices = [processor(x) for x in words]
        else:
            queries = sample
            choices = words

        def func():
            for query in queries:
                for choice in choices:
                    scorer(query, choice)

        return func

    fuzz = []
    rfuzz = []

    header_list = ["Function", "RapidFuzz", "FuzzyWuzzy", "SpeedImprovement"]
    row_format = "{:>25}" * len(header_list)
    print(row_format.format(*header_list))
    for target in LIBRARIES:
        scorer = load_func("fuzzywuzzy.fuzz." + target)
        sec = timeit(
            "func()",
            globals={"func": wrap_iterate(scorer, PROCESSOR[target])},
            number=1,
        )
        calls = total / sec
        fuzz.append(calls)

        rscorer = load_func("rapidfuzz.fuzz." + target)
        rsec = timeit("func()", globals={"func": wrap_cdist(rscorer, PROCESSOR[target])}, number=1)
        rcalls = total / rsec
        rfuzz.append(rcalls)

        print(row_format.format(target, f"{rcalls//1000}k", f"{calls//1000}k", f"{int(100 * sec/rsec)}%"))

    labels = LIBRARIES

    x = np.arange(len(labels))  # the label locations
    width = 0.35  # the width of the bars

    fig, ax = plt.subplots(figsize=(17, 10))
    rects1 = ax.bar(x - width / 2, fuzz, width, label="FuzzyWuzzy", color="xkcd:coral")
    rects2 = ax.bar(x + width / 2, rfuzz, width, label="RapidFuzz", color="#6495ED")

    # Add some text for labels, title and custom x-axis tick labels, etc.
    ax.set_ylabel("evaluated word pairs [inputs/s]")
    ax.set_xlabel("Scorer")
    ax.set_title("The number of word pairs evaluated per second\n(the larger the better)")
    ax.set_xticks(x)
    ax.set_xticklabels(labels, rotation=30)
    ax.get_yaxis().set_major_formatter(mpl.ticker.FuncFormatter(lambda x, _: format(int(x), ",")))
    ax.legend()

    def autolabel(rects):
        """Attach a text label above each bar in *rects*, displaying its height."""
        for rect in rects:
            height = rect.get_height()
            ax.annotate(
                f"{int(height):,}",
                xy=(rect.get_x() + rect.get_width() / 2, height),
                xytext=(0, 3),  # 3 points vertical offset
                textcoords="offset points",
                ha="center",
                va="bottom",
            )

    autolabel(rects1)
    autolabel(rects2)
    fig.tight_layout()
    plt.show()


if __name__ == "__main__":
    benchmark()



"""
This benchmark suite has the following goals:

- Help to detect performance regression in a new version
- Validate that an optimization change makes RapidFuzz faster and doesn't lead to major performance regressions
- Compare multiple implementations of the same algorithms in different libraries
- Showcase performance which ideally would be representative of performances of applications running
  on production
"""

from __future__ import annotations

import json
import math
import os
import random
import string
import subprocess
import sys
import timeit
from multiprocessing import Process, Value
from pathlib import Path
from urllib.request import urlopen

import click
import matplotlib.pyplot as plt
import pandas as pd
from packaging.version import Version
from tqdm import tqdm


def benchmark(result, setup, func, queries, choices):
    timer = timeit.Timer(func, setup=setup, globals={"queries": queries, "choices": choices})

    number = 1
    while True:
        timings = timer.repeat(repeat=7, number=number)
        min_timing = min(timings)
        # hopefully this reduces derivations in the benchmarks a bit
        if min_timing < 0.05:
            multiplicator = max(2, 0.05 / min_timing)
            number = math.ceil(number * multiplicator)
        else:
            result.value = min_timing / number
            return


# todo not sure how to handle ranges optimally here
# e.g. someone might want to
# - benchmark against all versions compatible
# - only the latest patch of each major
# - a custom range
# - only the installed version
def get_rapidfuzz_versions(version):
    url = "https://pypi.org/pypi/rapidfuzz/json"
    data = json.load(urlopen(url))
    versions = list(data["releases"].keys())
    versions.sort(key=Version, reverse=True)

    if version == "latest_patches":
        filtered_versions = []
        last_ver = None
        for version in versions:
            ver = Version(version)
            if last_ver is not None and last_ver.major == ver.major and last_ver.minor == ver.minor:
                continue

            last_ver = ver
            filtered_versions.append(ver)

        versions = filtered_versions

    return versions


def run_benchmark(datasets, setup, func) -> bool:
    results = []
    for dataset in tqdm(datasets["data"]):
        queries = [dataset[0]]
        choices = dataset[1:]

        result = Value("f", 0.0)
        # running this in a different process ensures, that imports do not "leak" out
        # this is relevant, since for regressions tests we install different versions of the module
        # during runtime
        process = Process(target=benchmark, args=(result, setup, func, queries, choices))
        process.start()
        process.join()
        # something went wrong
        if result.value == 0.0:
            return []
        results.append(result.value / len(choices))

    return results


def run_benchmarks_rapidfuzz(rapidfuzz_version, func_name, dataset, result_df):
    SCORERS = {
        "Jaro": "from rapidfuzz.distance.metrics_cpp import jaro_similarity as scorer",
        "JaroWinkler": "from rapidfuzz.distance.metrics_cpp import jaro_winkler_similarity as scorer",
        "OSA": "from rapidfuzz.distance.metrics_cpp import osa_distance as scorer",
        "Levenshtein": "from rapidfuzz.distance.metrics_cpp import levenshtein_distance as scorer",
        "Indel": "from rapidfuzz.distance.metrics_cpp import indel_distance as scorer",
        "DamerauLevenshtein": "from rapidfuzz.distance.metrics_cpp import damerau_levenshtein_distance as scorer",
        "Hamming": "from rapidfuzz.distance.metrics_cpp import hamming_distance as scorer",
    }

    setup = SCORERS[func_name]
    func = "[[scorer(a, b) for a in queries] for b in choices]"

    if rapidfuzz_version == "current":
        print("Benchmarking rapidfuzz")
        result_df["rapidfuzz"] = run_benchmark(dataset, setup, func)
    else:
        rapidfuzz_versions = get_rapidfuzz_versions(rapidfuzz_version)
        os.environ["RAPIDFUZZ_BUILD_EXTENSION"] = "1"

        for version in rapidfuzz_versions:
            print(f"Benchmarking rapidfuzz=={version}")
            res = subprocess.run(
                [sys.executable, "-m", "pip", "install", f"rapidfuzz=={version}"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=False,
            )
            # older versions are likely not installable
            if res.returncode:
                break

            # likely broken install on old version
            results = run_benchmark(dataset, setup, func)
            if not results:
                break

            result_df[f"rapidfuzz=={version}"] = results


def run_benchmarks_jellyfish(func_name, result_df, dataset):
    setup = "import jellyfish"
    JELLYFISH_SCORERS = {
        "Jaro": "jaro_similarity",
        "JaroWinkler": "jaro_winkler_similarity",
        "Levenshtein": "levenshtein_distance",
        "DamerauLevenshtein": "damerau_levenshtein_distance",
        "Hamming": "hamming_distance",
    }

    if func_name not in JELLYFISH_SCORERS:
        return

    print("Benchmarking jellyfish")
    func = f"[[jellyfish.{JELLYFISH_SCORERS[func_name]}(a, b) for a in queries] for b in choices]"
    result_df["jellyfish"] = run_benchmark(dataset, setup, func)


def run_benchmarks_polyleven(func_name, result_df, dataset):
    if func_name != "Levenshtein":
        return

    print("Benchmarking polyleven")
    setup = "import polyleven"
    func = "[[polyleven.levenshtein(a, b) for a in queries] for b in choices]"
    result_df["polyleven"] = run_benchmark(dataset, setup, func)


def run_benchmarks_edlib(func_name, result_df, dataset):
    if func_name != "Levenshtein":
        return

    print("Benchmarking edlib")
    setup = "import edlib"
    func = "[[edlib.align(a, b) for a in queries] for b in choices]"
    result_df["edlib"] = run_benchmark(dataset, setup, func)

    print("Benchmarking edlib(k=max)")
    func = "[[edlib.align(a, b, k=max(len(a), len(b))) for a in queries] for b in choices]"
    result_df["edlib(k=max)"] = run_benchmark(dataset, setup, func)


def run_benchmarks_editdistance(func_name, result_df, dataset):
    if func_name != "Levenshtein":
        return

    print("Benchmarking editdistance")
    setup = "import editdistance"
    func = "[[editdistance.eval(a, b) for a in queries] for b in choices]"
    result_df["editdistance"] = run_benchmark(dataset, setup, func)


def run_benchmarks_pyxdameraulevenshtein(func_name, result_df, dataset):
    if func_name != "OSA":
        return

    print("Benchmarking pyxdameraulevenshtein")
    setup = "import pyxdameraulevenshtein"
    func = "[[pyxdameraulevenshtein.damerau_levenshtein_distance(a, b) for a in queries] for b in choices]"
    result_df["pyxdameraulevenshtein"] = run_benchmark(dataset, setup, func)


AVAILABLE_BENCHMARKS = [
    "Jaro",
    "JaroWinkler",
    "OSA",
    "Levenshtein",
    "Indel",
    "DamerauLevenshtein",
]


@click.group
def cli():
    pass


@cli.command()
@click.option("--rapidfuzz_version", default="current", help="versions of rapidfuzz to benchmark")
@click.option(
    "--func",
    default="Levenshtein",
    type=click.Choice(AVAILABLE_BENCHMARKS, case_sensitive=False),
    help="function to benchmark",
)
def bench(rapidfuzz_version, func):
    temp_dir = Path(__file__).parent / "temp"
    temp_dir.mkdir(exist_ok=True)

    with open(temp_dir / "dataset.json") as f:
        dataset = json.load(f)

    result_df = pd.DataFrame(data={"x_axis": dataset["x_axis"]})

    run_benchmarks_rapidfuzz(rapidfuzz_version, func, dataset, result_df)

    # third party libs
    # run_benchmarks_jellyfish(func, result_df, dataset)
    run_benchmarks_polyleven(func, result_df, dataset)
    run_benchmarks_edlib(func, result_df, dataset)
    # run_benchmarks_editdistance(func, result_df, dataset)
    run_benchmarks_pyxdameraulevenshtein(func, result_df, dataset)

    result_df.to_csv(temp_dir / "result.csv", sep=",", index=False)


@cli.group()
def generate_dataset():
    pass


@generate_dataset.command()
@click.option("--start", default=0)
@click.option("--end", default=512)
@click.option("--step", default=4)
@click.option("--count", default=100)
def length_based(start, end, step, count):
    temp_dir = Path(__file__).parent / "temp"
    temp_dir.mkdir(exist_ok=True)

    random.seed(18)
    characters = string.ascii_letters + string.digits + string.whitespace + string.punctuation
    dataset = {"x_label": "string length [in characters]", "x_axis": [], "data": []}
    for i in range(start, end, step):
        data = ["".join(random.choice(characters) for _ in range(i)) for _ in range(count + 1)]
        dataset["data"].append(data)
        dataset["x_axis"].append(i)

    with open(temp_dir / "dataset.json", "w") as f:
        json.dump(dataset, f)


@cli.command()
def show():
    temp_dir = Path(__file__).parent / "temp"
    temp_dir.mkdir(exist_ok=True)

    with open(temp_dir / "dataset.json") as f:
        dataset = json.load(f)

    results = pd.read_csv("temp/result.csv")

    results.loc[:, (results.columns != "x_axis")] *= 1000 * 1000
    ax = results.plot(x="x_axis")

    # plt.xticks(list(range(0, 64*20+1, 64)))

    plt.title("Performance comparison of the \nDamerauLevenshtein similarity in different libraries")
    plt.xlabel(dataset["x_label"])
    plt.ylabel("runtime [μs]")
    ax.set_xlim(xmin=0)
    # ax.set_ylim(bottom=0)
    plt.grid()
    plt.show()


if __name__ == "__main__":
    cli()



# distutils: language=c++
# cython: language_level=3, binding=True, linetrace=True

from cpython.object cimport PyObject
from cpython.pycapsule cimport PyCapsule_GetPointer, PyCapsule_IsValid, PyCapsule_New
from libc.stddef cimport wchar_t
from libc.stdint cimport int64_t, uint64_t, SIZE_MAX, UINT64_MAX
from libc.stdlib cimport free, malloc
from libc.math cimport NAN
from libcpp cimport bool
from libcpp.cmath cimport isnan
from libcpp.utility cimport move, pair
from libcpp.vector cimport vector

from rapidfuzz cimport (
    PREPROCESSOR_STRUCT_VERSION,
    SCORER_STRUCT_VERSION,
    RF_GetScorerFlags,
    RF_Kwargs,
    RF_KwargsInit,
    RF_Preprocess,
    RF_Preprocessor,
    RF_Scorer,
    RF_ScorerFlags,
    RF_ScorerFuncInit,
    RF_UncachedScorerFunc,
    RF_String,
    RF_StringType,
)

# these are only for the cython compiler
# each module still has to define them on it's own
from array import array
pandas_NA = None

cdef extern from "rapidfuzz/details/types.hpp" namespace "rapidfuzz" nogil:
    cpdef enum class EditType:
        None    = 0,
        Replace = 1,
        Insert  = 2,
        Delete  = 3

    ctypedef struct RfEditOp "rapidfuzz::EditOp":
        EditType type
        size_t src_pos
        size_t dest_pos

    cdef cppclass RfScoreAlignment "rapidfuzz::ScoreAlignment"[T]:
        T score
        size_t src_start
        size_t src_end
        size_t dest_start
        size_t dest_end

        RfScoreAlignment()
        RfScoreAlignment(T score, size_t src_start, size_t src_end, size_t dest_start, size_t dest_end)

    cdef cppclass RfOpcodes "rapidfuzz::Opcodes"

    cdef cppclass RfEditops "rapidfuzz::Editops":
        ctypedef size_t size_type
        ctypedef ptrdiff_t difference_type

        cppclass iterator:
            iterator() except +
            iterator(iterator&) except +
            RfEditOp& operator*()
            iterator operator++()
            iterator operator--()
            iterator operator++(int)
            iterator operator--(int)
            iterator operator+(size_type)
            iterator operator-(size_type)
            difference_type operator-(iterator)
            difference_type operator-(const_iterator)
            bint operator==(iterator)
            bint operator==(const_iterator)
            bint operator!=(iterator)
            bint operator!=(const_iterator)
            bint operator<(iterator)
            bint operator<(const_iterator)
            bint operator>(iterator)
            bint operator>(const_iterator)
            bint operator<=(iterator)
            bint operator<=(const_iterator)
            bint operator>=(iterator)
            bint operator>=(const_iterator)
        cppclass const_iterator:
            const_iterator() except +
            const_iterator(iterator&) except +
            const_iterator(const_iterator&) except +
            operator=(iterator&) except +
            const RfEditOp& operator*()
            const_iterator operator++()
            const_iterator operator--()
            const_iterator operator++(int)
            const_iterator operator--(int)
            const_iterator operator+(size_type)
            const_iterator operator-(size_type)
            difference_type operator-(iterator)
            difference_type operator-(const_iterator)
            bint operator==(iterator)
            bint operator==(const_iterator)
            bint operator!=(iterator)
            bint operator!=(const_iterator)
            bint operator<(iterator)
            bint operator<(const_iterator)
            bint operator>(iterator)
            bint operator>(const_iterator)
            bint operator<=(iterator)
            bint operator<=(const_iterator)
            bint operator>=(iterator)
            bint operator>=(const_iterator)

        RfEditops() except +
        RfEditops(size_t) except +
        RfEditops(const RfEditops&) except +
        RfEditops(const RfOpcodes&) except +
        bool operator==(const RfEditops&)
        RfEditOp& operator[](size_t pos) except +
        size_t size()
        RfEditops inverse() except +
        RfEditops remove_subsequence(const RfEditops& subsequence) except +
        RfEditops slice(int, int, int) except +
        void remove_slice(int, int, int) except +
        size_t get_src_len()
        void set_src_len(size_t)
        size_t get_dest_len()
        void set_dest_len(size_t)
        RfEditops reverse()
        void emplace_back(...)
        void reserve(size_t) except +
        void shrink_to_fit() except +

        iterator begin()
        iterator end()
        const_iterator cbegin()
        const_iterator cend()
        iterator erase(iterator)

    ctypedef struct RfOpcode "rapidfuzz::Opcode":
        EditType type
        size_t src_begin
        size_t src_end
        size_t dest_begin
        size_t dest_end

    cdef cppclass RfOpcodes "rapidfuzz::Opcodes":
        ctypedef size_t size_type
        ctypedef ptrdiff_t difference_type

        cppclass iterator:
            iterator() except +
            iterator(iterator&) except +
            RfOpcode& operator*()
            iterator operator++()
            iterator operator--()
            iterator operator++(int)
            iterator operator--(int)
            iterator operator+(size_type)
            iterator operator-(size_type)
            difference_type operator-(iterator)
            difference_type operator-(const_iterator)
            bint operator==(iterator)
            bint operator==(const_iterator)
            bint operator!=(iterator)
            bint operator!=(const_iterator)
            bint operator<(iterator)
            bint operator<(const_iterator)
            bint operator>(iterator)
            bint operator>(const_iterator)
            bint operator<=(iterator)
            bint operator<=(const_iterator)
            bint operator>=(iterator)
            bint operator>=(const_iterator)
        cppclass const_iterator:
            const_iterator() except +
            const_iterator(iterator&) except +
            const_iterator(const_iterator&) except +
            operator=(iterator&) except +
            const RfOpcode& operator*()
            const_iterator operator++()
            const_iterator operator--()
            const_iterator operator++(int)
            const_iterator operator--(int)
            const_iterator operator+(size_type)
            const_iterator operator-(size_type)
            difference_type operator-(iterator)
            difference_type operator-(const_iterator)
            bint operator==(iterator)
            bint operator==(const_iterator)
            bint operator!=(iterator)
            bint operator!=(const_iterator)
            bint operator<(iterator)
            bint operator<(const_iterator)
            bint operator>(iterator)
            bint operator>(const_iterator)
            bint operator<=(iterator)
            bint operator<=(const_iterator)
            bint operator>=(iterator)
            bint operator>=(const_iterator)

        RfOpcodes() except +
        RfOpcodes(size_t) except +
        RfOpcodes(const RfOpcodes&) except +
        RfOpcodes(const RfEditops&) except +
        bool operator==(const RfOpcodes&)
        RfOpcode& operator[](size_t pos) except +
        size_t size()
        RfOpcodes inverse() except +
        size_t get_src_len()
        void set_src_len(size_t)
        size_t get_dest_len()
        void set_dest_len(size_t)
        RfOpcodes reverse()
        void emplace_back(...)
        void reserve(size_t) except +
        RfOpcode& back() except +
        void shrink_to_fit() except +
        bint empty()

        iterator begin()
        iterator end()
        const_iterator cbegin()
        const_iterator cend()
        iterator erase(iterator)

cdef extern from "cpp_common.hpp":
    cdef cppclass RF_StringWrapper:
        RF_String string
        PyObject* obj

        RF_StringWrapper()
        RF_StringWrapper(RF_String)
        RF_StringWrapper(RF_String, object)

    cdef cppclass RF_KwargsWrapper:
        RF_Kwargs kwargs

        RF_KwargsWrapper()
        RF_KwargsWrapper(RF_Kwargs)

    cdef cppclass PyObjectWrapper:
        PyObject* obj

        PyObjectWrapper()
        PyObjectWrapper(object)

    void default_string_deinit(RF_String* string) nogil

    int is_valid_string(object py_str) except +
    RF_String convert_string(object py_str)
    void validate_string(object py_str, const char* err) except +

cdef inline bool hash_array(arr, RF_String* s_proc) except False:
    # TODO on Cpython this does not require any copies
    cdef Py_UCS4 typecode = <Py_UCS4>arr.typecode
    s_proc.length = <int64_t>len(arr)

    s_proc.data = malloc(s_proc.length * sizeof(uint64_t))

    if s_proc.data == NULL:
        raise MemoryError

    try:
        # ignore signed/unsigned, since it is not relevant in any of the algorithms
        if typecode in {'b', 'B'}: # signed/unsigned char
            s_proc.kind = RF_StringType.RF_UINT64
            for i in range(s_proc.length):
                (<uint64_t*>s_proc.data)[i] = <uint64_t>arr[i]
        elif typecode == 'u': # 'u' wchar_t
            s_proc.kind = RF_StringType.RF_UINT64
            for i in range(s_proc.length):
                (<uint64_t*>s_proc.data)[i] = <uint64_t><Py_UCS4>arr[i]
        elif typecode in {'h', 'H'}: #  signed/unsigned short
            s_proc.kind = RF_StringType.RF_UINT64
            for i in range(s_proc.length):
                (<uint64_t*>s_proc.data)[i] = <uint64_t>arr[i]
        elif typecode in {'i', 'I'}: # signed/unsigned int
            s_proc.kind = RF_StringType.RF_UINT64
            for i in range(s_proc.length):
                (<uint64_t*>s_proc.data)[i] = <uint64_t>arr[i]
        elif typecode in {'l', 'L'}: # signed/unsigned long
            s_proc.kind = RF_StringType.RF_UINT64
            for i in range(s_proc.length):
                (<uint64_t*>s_proc.data)[i] = <uint64_t>arr[i]
        elif typecode in {'q', 'Q'}: # signed/unsigned long long
            s_proc.kind = RF_StringType.RF_UINT64
            for i in range(s_proc.length):
                (<uint64_t*>s_proc.data)[i] = <uint64_t>arr[i]
        else: # float/double are hashed
            s_proc.kind = RF_StringType.RF_UINT64
            for i in range(s_proc.length):
                (<uint64_t*>s_proc.data)[i] = <uint64_t>hash(arr[i])
    except Exception as e:
        free(s_proc.data)
        s_proc.data = NULL
        raise

    s_proc.dtor = default_string_deinit
    return True


cdef inline bool hash_sequence(seq, RF_String* s_proc) except False:
    s_proc.length = <int64_t>len(seq)

    s_proc.data = malloc(s_proc.length * sizeof(uint64_t))

    if s_proc.data == NULL:
        raise MemoryError

    try:
        s_proc.kind = RF_StringType.RF_UINT64
        for i in range(s_proc.length):
            elem = seq[i]
            # this is required so e.g. a list of char can be compared to a string
            if isinstance(elem, str) and len(elem) == 1:
                (<uint64_t*>s_proc.data)[i] = <uint64_t><Py_UCS4>elem
            elif isinstance(elem, int) and elem == -1:
                (<uint64_t*>s_proc.data)[i] = <uint64_t>-1
            else:
                (<uint64_t*>s_proc.data)[i] = <uint64_t>hash(elem)
    except Exception as e:
        free(s_proc.data)
        s_proc.data = NULL
        raise

    s_proc.dtor = default_string_deinit
    return True

cdef inline bool is_none(s) noexcept:
    if s is None or s is pandas_NA:
        return True

    if isinstance(s, float) and isnan(<double>s):
        return True

    return False

# todo we will probably want to clean up the various methods of
# converting strings. This has to be done carefully, since especially with preprocessor functions
# the none check is often required before calling the preprocessing functions to keep the current behaviour
cdef inline bool conv_sequence_with_none(seq, RF_String* c_seq) except False:
    if is_valid_string(seq):
        c_seq[0] = move(convert_string(seq))
    elif is_none(seq):
        c_seq.length = 0
        c_seq.data = NULL
    elif isinstance(seq, array):
        hash_array(seq, c_seq)
    else:
        hash_sequence(seq, c_seq)

    return True

cdef inline RF_String conv_sequence(seq) except *:
    cdef RF_String c_seq
    if is_valid_string(seq):
        c_seq = move(convert_string(seq))
    elif isinstance(seq, array):
        hash_array(seq, &c_seq)
    else:
        hash_sequence(seq, &c_seq)

    return move(c_seq)

cdef inline double get_score_cutoff_f64(score_cutoff, float worst_score, float optimal_score) except *:
    cdef float c_score_cutoff = worst_score

    if score_cutoff is not None:
        c_score_cutoff = score_cutoff
        if optimal_score > worst_score:
            # e.g. 0.0 - 100.0
            if c_score_cutoff < worst_score or c_score_cutoff > optimal_score:
                raise TypeError("score_cutoff has to be in the range of %s - %s" % (worst_score, optimal_score))
        else:
            # e.g. DBL_MAX - 0
            if c_score_cutoff > worst_score or c_score_cutoff < optimal_score:
                raise TypeError(f"score_cutoff has to be in the range of {optimal_score} - {worst_score}")

    return c_score_cutoff

cdef inline int64_t get_score_cutoff_i64(score_cutoff, int64_t worst_score, int64_t optimal_score) except *:
    cdef int64_t c_score_cutoff = worst_score

    if score_cutoff is not None:
        if optimal_score > worst_score:
            # e.g. 0.0 - 100.0
            if c_score_cutoff < worst_score or c_score_cutoff > optimal_score:
                raise TypeError(f"score_cutoff has to be in the range of {worst_score} - {optimal_score}")
        else:
            # e.g. DBL_MAX - 0
            if c_score_cutoff > worst_score or c_score_cutoff < optimal_score:
                raise TypeError(f"score_cutoff has to be in the range of {optimal_score} - {worst_score}")

    return c_score_cutoff

cdef inline size_t get_score_cutoff_size_t(score_cutoff, size_t worst_score, size_t optimal_score) except *:
    cdef uint64_t c_score_cutoff = worst_score

    if score_cutoff is not None:
        c_score_cutoff = score_cutoff
        if c_score_cutoff > SIZE_MAX:
            c_score_cutoff = SIZE_MAX

        if optimal_score > worst_score:
            # e.g. 0.0 - 100.0
            if c_score_cutoff < worst_score or c_score_cutoff > optimal_score:
                raise TypeError(f"score_cutoff has to be in the range of {worst_score} - {optimal_score}")
        else:
            # e.g. DBL_MAX - 0
            if c_score_cutoff > worst_score or c_score_cutoff < optimal_score:
                raise TypeError(f"score_cutoff has to be in the range of {optimal_score} - {worst_score}")

    return <size_t>c_score_cutoff

cdef inline bool preprocess_strings(s1, s2, processor, RF_StringWrapper* s1_proc, RF_StringWrapper* s2_proc) except False:
    cdef RF_Preprocessor* preprocess_context = NULL

    if not processor:
        s1_proc[0] = RF_StringWrapper(conv_sequence(s1))
        s2_proc[0] = RF_StringWrapper(conv_sequence(s2))
    else:
        processor_capsule = getattr(processor, '_RF_Preprocess', processor)
        if PyCapsule_IsValid(processor_capsule, NULL):
            preprocess_context = <RF_Preprocessor*>PyCapsule_GetPointer(processor_capsule, NULL)

        if preprocess_context != NULL and preprocess_context.version == 1:
            preprocess_context.preprocess(s1, &(s1_proc[0].string))
            preprocess_context.preprocess(s2, &(s2_proc[0].string))
        else:
            s1 = processor(s1)
            s1_proc[0] = RF_StringWrapper(conv_sequence(s1), s1)
            s2 = processor(s2)
            s2_proc[0] = RF_StringWrapper(conv_sequence(s2), s2)

    return True

cdef inline bool NoKwargsInit(RF_Kwargs* self, dict kwargs) except False:
    if len(kwargs):
        raise TypeError("Got unexpected keyword arguments: ", ", ".join(kwargs.keys()))

    self.context = NULL
    self.dtor = NULL
    return True

cdef inline RF_Scorer CreateScorerContext(RF_KwargsInit kwargs_init, RF_GetScorerFlags get_scorer_flags, RF_ScorerFuncInit scorer_func_init, RF_UncachedScorerFunc uncached_scorer_func) noexcept:
    cdef RF_Scorer context
    context.version = SCORER_STRUCT_VERSION
    context.kwargs_init = kwargs_init
    context.get_scorer_flags = get_scorer_flags
    context.scorer_func_init = scorer_func_init
    context.uncached_scorer_func = uncached_scorer_func
    return context

cdef inline void SetFuncAttrs(cpp_func, py_func) except *:
    cpp_func.__name__ = py_func.__name__
    cpp_func.__qualname__ = py_func.__qualname__
    cpp_func.__doc__ = py_func.__doc__

cdef inline void SetScorerAttrs(cpp_func, py_func, RF_Scorer* context) except *:
    SetFuncAttrs(cpp_func, py_func)
    cpp_func._RF_Scorer = PyCapsule_New(context, NULL, NULL)
    cpp_func._RF_ScorerPy = py_func._RF_ScorerPy

    # used to detect the function hasn't been wrapped afterwards
    cpp_func._RF_OriginalScorer = cpp_func

cdef inline RF_Preprocessor CreateProcessorContext(RF_Preprocess preprocess) except *:
    cdef RF_Preprocessor context
    context.version = PREPROCESSOR_STRUCT_VERSION
    context.preprocess = preprocess
    return context

cdef inline void SetProcessorAttrs(cpp_func, py_func, RF_Preprocessor* context) except *:
    SetFuncAttrs(cpp_func, py_func)
    cpp_func._RF_Preprocess = PyCapsule_New(context, NULL, NULL)



from __future__ import annotations

from pathlib import Path

from git import Repo, cmd


def lsremote(url):
    """
    only one way for
    git ls-remote -t --exit-code --refs {url} | sed -E 's/^[[:xdigit:]]+[[:space:]]+refs\\/tags\\/(.+)/\1/g'
    """
    g = cmd.Git()
    refs = g.ls_remote("-t", "--exit-code", url).split("\n")
    return refs[-1].split("\t")[-1].split("/")[-1]


def test_submodules():
    """
    find tag in current repo and check tag in remote submodule
    """
    repo_path = Path(__file__).resolve().parent.parent
    repo = Repo(repo_path)  # init class Repo
    assert not repo.bare  # checking if repo is bare

    for module in repo.submodules:
        sub_repo = module.module()  # init class Repo for submodule
        assert not sub_repo.bare
        current_tag = next(
            (tag for tag in sub_repo.tags if tag.commit == sub_repo.head.commit), None
        )  # find current tag in local repo

        assert current_tag is not None, f"You don't use tag for '{module.name}'!"

        current_tag = current_tag.name  # git.Tag to str
        remote_tag = lsremote(module.url)  # find last tag in remote module

        assert current_tag == remote_tag, f"Local submodule tag {current_tag} does not match remote {remote_tag}"


if __name__ == "__main__":
    test_submodules()



from __future__ import annotations

from rapidfuzz._feature_detector import AVX2, supports
from rapidfuzz.distance import metrics_cpp, metrics_py
from tests.common import GenericScorer, Scorer, is_none

cpp_scorer_modules = [metrics_cpp]
try:
    if supports(AVX2):
        from rapidfuzz.distance import metrics_cpp_avx2

        cpp_scorer_modules.append(metrics_cpp_avx2)
except Exception:
    pass


def create_generic_scorer(func_name, get_scorer_flags):
    py_scorers = [
        Scorer(
            distance=getattr(metrics_py, func_name + "_distance"),
            similarity=getattr(metrics_py, func_name + "_similarity"),
            normalized_distance=getattr(metrics_py, func_name + "_normalized_distance"),
            normalized_similarity=getattr(metrics_py, func_name + "_normalized_similarity"),
            editops=getattr(metrics_py, func_name + "_editops", None),
            opcodes=getattr(metrics_py, func_name + "_opcodes", None),
        )
    ]

    cpp_scorers = [
        Scorer(
            distance=getattr(mod, func_name + "_distance"),
            similarity=getattr(mod, func_name + "_similarity"),
            normalized_distance=getattr(mod, func_name + "_normalized_distance"),
            normalized_similarity=getattr(mod, func_name + "_normalized_similarity"),
            editops=getattr(metrics_cpp, func_name + "_editops", None),
            opcodes=getattr(metrics_cpp, func_name + "_opcodes", None),
        )
        for mod in cpp_scorer_modules
    ]

    return GenericScorer(py_scorers, cpp_scorers, get_scorer_flags)


def get_scorer_flags_damerau_levenshtein(s1, s2, **kwargs):
    if is_none(s1) or is_none(s2):
        return {"maximum": None, "symmetric": True}
    return {"maximum": max(len(s1), len(s2)), "symmetric": True}


DamerauLevenshtein = create_generic_scorer("damerau_levenshtein", get_scorer_flags_damerau_levenshtein)


def get_scorer_flags_hamming(s1, s2, **kwargs):
    if is_none(s1) or is_none(s2):
        return {"maximum": None, "symmetric": True}
    return {"maximum": max(len(s1), len(s2)), "symmetric": True}


Hamming = create_generic_scorer("hamming", get_scorer_flags_hamming)


def get_scorer_flags_indel(s1, s2, **kwargs):
    if is_none(s1) or is_none(s2):
        return {"maximum": None, "symmetric": True}
    return {"maximum": len(s1) + len(s2), "symmetric": True}


Indel = create_generic_scorer("indel", get_scorer_flags_indel)


def get_scorer_flags_jaro(s1, s2, **kwargs):
    if is_none(s1) or is_none(s2):
        return {"maximum": None, "symmetric": True}
    return {"maximum": 1.0, "symmetric": True}


Jaro = create_generic_scorer("jaro", get_scorer_flags_jaro)


def get_scorer_flags_jaro_winkler(s1, s2, **kwargs):
    if is_none(s1) or is_none(s2):
        return {"maximum": None, "symmetric": True}
    return {"maximum": 1.0, "symmetric": True}


JaroWinkler = create_generic_scorer("jaro_winkler", get_scorer_flags_jaro_winkler)


def get_scorer_flags_lcs_seq(s1, s2, **kwargs):
    if is_none(s1) or is_none(s2):
        return {"maximum": None, "symmetric": True}
    return {"maximum": max(len(s1), len(s2)), "symmetric": True}


LCSseq = create_generic_scorer("lcs_seq", get_scorer_flags_lcs_seq)


def get_scorer_flags_levenshtein(s1, s2, weights=(1, 1, 1), **kwargs):
    insert_cost, delete_cost, replace_cost = weights

    if is_none(s1) or is_none(s2):
        return {"maximum": None, "symmetric": insert_cost == delete_cost}

    max_dist = len(s1) * delete_cost + len(s2) * insert_cost

    if len(s1) >= len(s2):
        max_dist = min(max_dist, len(s2) * replace_cost + (len(s1) - len(s2)) * delete_cost)
    else:
        max_dist = min(max_dist, len(s1) * replace_cost + (len(s2) - len(s1)) * insert_cost)

    return {"maximum": max_dist, "symmetric": insert_cost == delete_cost}


Levenshtein = create_generic_scorer("levenshtein", get_scorer_flags_levenshtein)


def get_scorer_flags_osa(s1, s2, **kwargs):
    if is_none(s1) or is_none(s2):
        return {"maximum": None, "symmetric": True}
    return {"maximum": max(len(s1), len(s2)), "symmetric": True}


OSA = create_generic_scorer("osa", get_scorer_flags_osa)


def get_scorer_flags_postfix(s1, s2, **kwargs):
    if is_none(s1) or is_none(s2):
        return {"maximum": None, "symmetric": True}
    return {"maximum": max(len(s1), len(s2)), "symmetric": True}


Postfix = create_generic_scorer("postfix", get_scorer_flags_postfix)


def get_scorer_flags_prefix(s1, s2, **kwargs):
    if is_none(s1) or is_none(s2):
        return {"maximum": None, "symmetric": True}
    return {"maximum": max(len(s1), len(s2)), "symmetric": True}


Prefix = create_generic_scorer("prefix", get_scorer_flags_prefix)

all_scorer_modules = [
    DamerauLevenshtein,
    Hamming,
    Indel,
    Jaro,
    JaroWinkler,
    LCSseq,
    Levenshtein,
    OSA,
    Postfix,
    Prefix,
]



# SPDX-License-Identifier: MIT
# Copyright (C) 2025 Max Bachmann
# This file is generated by tools/generate_python.py
from __future__ import annotations

import contextlib
import os

from rapidfuzz._feature_detector import AVX2, SSE2, supports

__all__ = ["cdist", "cpdist", "extract", "extractOne", "extract_iter"]

_impl = os.environ.get("RAPIDFUZZ_IMPLEMENTATION")
if _impl == "cpp":
    imported = False
    if supports(AVX2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.process_cpp_avx2 import (  # pyright: ignore[reportMissingImports]
                cdist,
                cpdist,
                extract,
                extract_iter,
                extractOne,
            )

            imported = True

    if not imported and supports(SSE2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.process_cpp_sse2 import (  # pyright: ignore[reportMissingImports]
                cdist,
                cpdist,
                extract,
                extract_iter,
                extractOne,
            )

            imported = True

    if not imported:
        from rapidfuzz.process_cpp import (  # pyright: ignore[reportMissingImports]
            cdist,
            cpdist,
            extract,
            extract_iter,
            extractOne,
        )
elif _impl == "python":
    from rapidfuzz.process_py import cdist, cpdist, extract, extract_iter, extractOne
else:
    imported = False
    if supports(AVX2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.process_cpp_avx2 import (  # pyright: ignore[reportMissingImports]
                cdist,
                cpdist,
                extract,
                extract_iter,
                extractOne,
            )

            imported = True

    if not imported and supports(SSE2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.process_cpp_sse2 import (  # pyright: ignore[reportMissingImports]
                cdist,
                cpdist,
                extract,
                extract_iter,
                extractOne,
            )

            imported = True

    if not imported:
        with contextlib.suppress(ImportError):
            from rapidfuzz.process_cpp import (  # pyright: ignore[reportMissingImports]
                cdist,
                cpdist,
                extract,
                extract_iter,
                extractOne,
            )

            imported = True

    if not imported:
        from rapidfuzz.process_py import (
            cdist,
            cpdist,
            extract,
            extract_iter,
            extractOne,
        )
